@@ -3,6 +3,7 @@ import cv2
 import yaml
 import boxx
 import numpy as np
+from functools import wraps
 
 with boxx.inpkg():
     from . import utils
@@ -39,11 +40,49 @@ class Stereo:
         self.get_R_t_by_stereo_calibrate()
         self._get_undistort_rectify_map()
 
+    def get_conjoint_points(self):
+        image_points = list(self.cam1.values())[0]["image_points"]
+        # "image_points" is checkboard type: shape(n, 2)
+        if isinstance(image_points, np.ndarray):
+            keys = self.cam1.valid_keys_intersection(self.cam2)
+            conjoint_image_points1 = [self.cam1[key]["image_points"] for key in keys]
+            conjoint_image_points2 = [self.cam2[key]["image_points"] for key in keys]
+            conjoint_object_points = [self.cam1[key]["object_points"] for key in keys]
+
+        # "image_points" is marker type: {id: shape(n, 2)}
+        elif isinstance(image_points, dict):
+            conjoint_image_points1, conjoint_image_points2, conjoint_object_points = (
+                [],
+                [],
+                [],
+            )
+            for key in self.cam1:
+                a, b, c = [], [], []
+                for id in self.cam1[key].get("image_points", {}):
+                    if key in self.cam2 and id in self.cam2[key].get(
+                        "image_points", {}
+                    ):
+                        a.append(self.cam1[key]["image_points"][id])
+                        b.append(self.cam2[key]["image_points"][id])
+                        c.append(self.cam1[key]["object_points"][id])
+                if a:
+                    conjoint_image_points1.append(np.concatenate(a, 0))
+                    conjoint_image_points2.append(np.concatenate(b, 0))
+                    conjoint_object_points.append(np.concatenate(c, 0))
+        return (
+            conjoint_image_points1,
+            conjoint_image_points2,
+            conjoint_object_points,
+        )
+
     def get_R_t_by_stereo_calibrate(self):
-        keys = self.cam1.valid_keys_intersection(self.cam2)
-        image_points1 = [self.cam1[key]["image_points"] for key in keys]
-        image_points2 = [self.cam2[key]["image_points"] for key in keys]
-        object_points = [self.cam1.object_points] * len(keys)
+
+        (
+            conjoint_image_points1,
+            conjoint_image_points2,
+            conjoint_object_points,
+        ) = self.get_conjoint_points()
+
         if self.force_same_intrinsic:
             self.cam2 = self.cam2.copy()
             self.cam2.K = self.cam1.K
@@ -58,9 +97,9 @@ class Stereo:
             1e-5,
         )
         ret, _, _, _, _, self.R, self.t, E, F = cv2.stereoCalibrate(
-            object_points,
-            image_points1,
-            image_points2,
+            conjoint_object_points,
+            conjoint_image_points1,
+            conjoint_image_points2,
             self.cam1.K,
             self.cam1.D,
             self.cam2.K,
@@ -147,7 +186,16 @@ class Stereo:
             return cv2.imread(path_or_np)[..., ::-1]
         return path_or_np
 
-    vis = staticmethod(utils.vis_stereo)
+    @wraps(utils.vis_stereo)
+    def vis(self, *args, **argkws):
+        if not isinstance(self, Stereo):
+            args = (self,) + args
+        if not args:
+            key = list(self.cam1)[0]
+            img1 = boxx.imread(self.cam1[key]["path"])
+            img2 = boxx.imread(self.cam2[key]["path"])
+            args = self.rectify(img1, img2)
+        return utils.vis_stereo(*args, **argkws)
 
     @classmethod
     def shows(cls, *l, **kv):
@@ -167,9 +215,7 @@ class Stereo:
         print(
             "-" * 15, "Stereo.precision_analysis", "-" * 15,
         )
-        print("Baseline: %.2fcm" % (100 * baseline))
-        print("cam1.xy:", self.cam1.xy)
-        print("cam1.fovs:", utils._str_angle_dic(self.cam1.fovs))
+        print(self)
         print("\n")
         dispmin = baseline * K[0, 0] / zmax
         dispmax = baseline * K[0, 0] / zmin
@@ -196,6 +242,15 @@ class Stereo:
             "-" * 15, "End of Stereo.precision_analysis", "-" * 15,
         )
 
+    def __str__(self):
+        strr = "Stereo:\n"
+        strr += "\tBaseline: %.2fcm\n" % (100 * self.baseline)
+        strr += "\tcam1.xy: %s\n" % str(self.cam1.xy)
+        strr += "\tcam1.fovs: %s" % str(utils._str_angle_dic(self.cam1.fovs))
+        return strr
+
+    __repr__ = __str__
+
     MAX_DEPTH = 10
 
     def get_max_depth(self):
@@ -339,26 +394,42 @@ class MetaStereoMatching:
 
 # A Example of StereoMatching class
 class SemiGlobalBlockMatching(MetaStereoMatching):
-    def __init__(self, cfg):
-        default_cfg = {
-            "max_size": 1000,
-            # SGBM_args
-            "P1": 4590,
-            "P2": 18360,
-            "blockSize": 5,
-            "disp12MaxDiff": 1,
-            "minDisparity": 2,
-            "mode": 2,
-            "numDisparities": 80,
-            "preFilterCap": 63,
-            "speckleRange": 2,
-            "speckleWindowSize": 0,
-            "uniquenessRatio": 15,
-        }
-        default_cfg.update(cfg)
-        self.cfg = default_cfg
-        self.max_size = self.cfg.pop("max_size")
-        self.stereo_sgbm = cv2.StereoSGBM_create(**self.cfg)
+    def __init__(self, cfg=None):
+        if cfg is None:
+            cfg = {}
+        self.cfg = cfg
+        self.max_size = self.cfg.get("max_size", 1000)
+
+        # StereoSGBM_create from https://gist.github.com/andijakl/ffe6e5e16742455291ef2a4edbe63cb7
+        block_size = 11
+        min_disp = 2
+        max_disp = 220
+        # Maximum disparity minus minimum disparity. The value is always greater than zero.
+        # In the current implementation, this parameter must be divisible by 16.
+        num_disp = max_disp - min_disp
+        # Margin in percentage by which the best (minimum) computed cost function value should "win" the second best value to consider the found match correct.
+        # Normally, a value within the 5-15 range is good enough
+        uniquenessRatio = 5
+        # Maximum size of smooth disparity regions to consider their noise speckles and invalidate.
+        # Set it to 0 to disable speckle filtering. Otherwise, set it somewhere in the 50-200 range.
+        speckleWindowSize = 200
+        # Maximum disparity variation within each connected component.
+        # If you do speckle filtering, set the parameter to a positive value, it will be implicitly multiplied by 16.
+        # Normally, 1 or 2 is good enough.
+        speckleRange = 2
+        disp12MaxDiff = 0
+
+        self.stereo_sgbm = cv2.StereoSGBM_create(
+            minDisparity=min_disp,
+            numDisparities=num_disp,
+            blockSize=block_size,
+            uniquenessRatio=uniquenessRatio,
+            speckleWindowSize=speckleWindowSize,
+            speckleRange=speckleRange,
+            disp12MaxDiff=disp12MaxDiff,
+            P1=8 * 1 * block_size * block_size,
+            P2=32 * 1 * block_size * block_size,
+        )
 
     def __call__(self, img1, img2):
         resize_ratio = min(self.max_size / max(img1.shape[:2]), 1)
@@ -366,7 +437,9 @@ class SemiGlobalBlockMatching(MetaStereoMatching):
         sdisparity = (self.stereo_sgbm.compute(simg1, simg2).astype(np.float32)).clip(
             0
         ) / 16.0
-        disparity = boxx.resize(sdisparity, 1 / resize_ratio) / resize_ratio
+        disparity = (
+            boxx.resize(sdisparity, img1.shape[:2]) * img1.shape[1] / simg1.shape[1]
+        )
         return disparity
 
 
@@ -377,6 +450,7 @@ if __name__ == "__main__":
     cam1, cam2, camd = Cam.get_test_cams()
 
     stereo = Stereo(cam1, cam2)
+    print(stereo)
 
     yaml_path = "/tmp/stereo.yaml"
     stereo.dump(yaml_path)
